﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Concurrency.TestTools.UnitTesting;
using System.Threading;

namespace AlpacaProject
{
    public class LambdasAndDelegates
    {

        private event EventHandler MyEvent;
        private void OnMyEvent()
        {
            if (MyEvent != null)
                MyEvent(this, EventArgs.Empty);
        }

        [UnitTestMethod]
        public void LambdaExamples()
        {
            // A lambda expression itself doesn't have a type.
            //var myLambda = x => x + 1; // CS0815

            Action ac = (() => Console.WriteLine());

            // Simple lambda expression
            Func<int, bool> isEqualToFive = x => x == 5;
            Assert.AreEqual(true, isEqualToFive(5));
            Assert.AreEqual(false, isEqualToFive(0));
            Assert.AreEqual(false, isEqualToFive(-5));

            Func<int, bool> isEven = x => x % 2 == 0;
            Assert.AreEqual(true, isEven(0));
            Assert.AreEqual(false, isEven(1));
            Assert.AreEqual(true, isEven(2));
            Assert.AreEqual(false, isEven(3));
            Assert.AreEqual(true, isEven(4));
            Assert.AreEqual(false, isEven(5));
            Assert.AreEqual(false, isEven(-1));
            Assert.AreEqual(true, isEven(-2));

            // Lambda w/o any args
            Func<double> getSecondsSinceMidnight = () =>
                (DateTime.Now - DateTime.Today).TotalSeconds;

            // You can explicitly specify the types of the args in a lambda
            Func<string, int, bool> isLongerThan = (string s, int x) =>
                s != null && s.Length > x;
            // But most times it's implicitly specified by the complier
            isLongerThan = (s, x) => s != null && s.Length > x;
            Assert.AreEqual(false, isLongerThan(null, 10));
            Assert.AreEqual(false, isLongerThan(String.Empty, 0));
            Assert.AreEqual(true, isLongerThan(" ", 0));
            Assert.AreEqual(true, isLongerThan("PPCP", 0));
            Assert.AreEqual(false, isLongerThan(String.Empty, 1));
            Assert.AreEqual(false, isLongerThan(" ", 1));
            Assert.AreEqual(true, isLongerThan("PPCP", 1));
            Assert.AreEqual(false, isLongerThan(String.Empty, 5));
            Assert.AreEqual(false, isLongerThan(" ", 5));
            Assert.AreEqual(false, isLongerThan("PPCP", 5));
            Assert.AreEqual(false, isLongerThan("12345", 5));
            Assert.AreEqual(true, isLongerThan("123456", 5));
            Assert.AreEqual(true, isLongerThan("1234567890", 5));

            // When you want to include intermmediate statements, rather than
            // a single expression just use the block syntax:
            Action<int> printNumbers = x => {
                for (int i = 0; i < x; i++)
                    Console.Write((i + 1) + " ");
                Console.WriteLine();
            };
            printNumbers(5);
            printNumbers(12);
        }

        [UnitTestMethod]
        public void LambdasOnListOfNumbers()
        {
            Random rand = new Random(0);    // Keep same for all tests

            int n = 1000;
            int maxValue = 20;
            List<int> nums = new List<int>(n);
            for (int i = 0; i < n; i++)
                nums.Add(rand.Next(maxValue));

            // 
            Assert.IsTrue(nums.TrueForAll(x => x >= 0), "Random.Next doesn't work as documented.");
            Assert.IsTrue(nums.TrueForAll(x => x < maxValue), "Random.Next doesn't work as documented.");

            int count3s = nums.FindAll(x => x == 3).Count;
            Assert.IsTrue(count3s != 0, "No threes exist in data; Pick a different random seed.");

            // Delete all the threes
            nums.RemoveAll(x => x == 3);
            Assert.IsTrue(nums.FindIndex(x => x == 3) == -1, "Some threes didn't get removed");
            Assert.AreEqual(n - count3s, nums.Count);
        }


        [UnitTestMethod]
        public void LambdaClosureExample()
        {
            // Both of these vars would normally be stored on the stack since
            // they are value types (structs) but since they are used in a
            // lambda expression a closure is created with them by boxing them
            // (placed on the heap).
            DateTime initTime = DateTime.Now;
            int tickCount = 0;

            // Since this var isn't accessed in a lambda, no closure is created
            // and thus can be stored on the stack.
            DateTime origInitTime = initTime;

            Action onTick = () => {
                // The closure
                tickCount++;

                TimeSpan timeSinceInit = DateTime.Now - initTime;
                Console.WriteLine(
                    "Time since function initialized: {0:m\\:s\\:FFF}"
                    , timeSinceInit
                    );
            };
            Assert.AreEqual(0, tickCount);
            Assert.AreEqual(origInitTime, initTime);
            tickCount++;

            onTick();
            Assert.AreEqual(1, tickCount);
            Assert.AreEqual(origInitTime, initTime);
            Thread.Sleep(1000 * 2);
            onTick();
            Assert.AreEqual(2, tickCount);
            Assert.AreEqual(origInitTime, initTime);
        }

        private Action CreateTick()
        {
            DateTime initTime = DateTime.Now;
            int tickCount = 0;

            // Since this var isn't accessed in a lambda, no closure is created
            // and thus can be stored on the stack.
            DateTime origInitTime = initTime;

            Action onTick = () => {
                // The closure
                tickCount++;

                TimeSpan timeSinceInit = DateTime.Now - initTime;
                Console.WriteLine(
                    "Tick: {0} - Time since function initialized: {1:m\\:s\\:FFF}"
                    , tickCount
                    , timeSinceInit
                    );
            };
            Assert.AreEqual(0, tickCount);
            Assert.AreEqual(origInitTime, initTime);

            return onTick;
        }

        [UnitTestMethod]
        public void LambdaClosureExample2()
        {
            Action onTick = CreateTick();

            onTick();
            Thread.Sleep(1000 * 2);
            onTick();
        }

        public void ConvertLambdaToDelegate()
        {
            var implicitDelegateNoArg = new Thread(() => { });
            var implicitDelegateWithArg = new Thread((obj) => { });

            // Using strong-typed delegates
            ThreadStart tstart1 = delegate() { };
            ThreadStart tstart2 = () => { };

            // Which of the following do you like better?
            // Handler w/annonymous method
            MyEvent += new EventHandler(delegate(object sender, EventArgs args) { });
            // Handler w/lambda
            MyEvent += new EventHandler((sender, args) => { });
            // Annonymous method
            MyEvent += delegate(object sender, EventArgs args) { };
            // Lambda only
            MyEvent += (sender, args) => { };
        }

    }
}
