﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Concurrency.TestTools.UnitTesting;
using System.Threading.Tasks;
using System.Threading;

namespace AlpacaProject
{
    public class TaskExamples
    {

        [UnitTestMethod]
        public void SimpleNestedTask()
        {
            var parent = Task.Factory.StartNew(() => {
                Console.WriteLine("Outer task executing.");
                var child = Task.Factory.StartNew(() => {
                    Console.WriteLine("Nested task starting.");
                    Thread.SpinWait(500000);
                    Console.WriteLine("Nested task completing.");
                });
            });
            parent.Wait();
            Console.WriteLine("Outer has completed.");
        }

        [UnitTestMethod]
        public void SimpleNestedTask_AttachedToParent()
        {
            var parent = Task.Factory.StartNew(() => {
                Console.WriteLine("Outer task executing.");
                var child = Task.Factory.StartNew(() => {
                    Console.WriteLine("Nested task starting.");
                    Thread.SpinWait(500000);
                    Console.WriteLine("Nested task completing.");
                }
                , TaskCreationOptions.AttachedToParent
                );
            });
            parent.Wait();
            Console.WriteLine("Outer has completed.");
        }

        [UnitTestMethod]
        public void TaskResults()
        {
            Task<int> dataForTheFuture = Task.Factory.StartNew(() => ComputeSomeResult(2000));

            // Do some other work while allowing for dataForTheFuture to compute asynchronously
            Thread.Sleep(100);

            // This will return the value immediately if the Task has already 
            // completed, or will wait for the result to be available if it’s 
            // not yet completed.
            DateTime start = DateTime.Now;
            int result = dataForTheFuture.Result;
            Console.WriteLine("result={0} (wait time={1})", result, (DateTime.Now - start));
        }

        [UnitTestMethod]
        public void TaskCancellation()
        {
            var cts = new CancellationTokenSource();
            Task<int> dataForTheFuture = Task.Factory.StartNew(() => ComputeEx(10000), cts.Token);

            // Do some other work while allowing for dataForTheFuture to compute asynchronously
            Thread.Sleep(1500);

            // Cancel it
            DateTime start = DateTime.Now;
            cts.Cancel();
            Console.WriteLine("cts.Cancel() finished (wait time={0})", (DateTime.Now - start));
        }

        [UnitTestMethod]
        public void TaskExceptionContinuation()
        {
            var cts = new CancellationTokenSource();
            Task<int> dataForTheFuture = Task.Factory.StartNew(() => ComputeEx(3000), cts.Token);

            // Do some other work while allowing for dataForTheFuture to compute asynchronously
            Thread.Sleep(1500);

            // Cancel it and make sure we are made aware of any exceptions
            // that occurred.
            dataForTheFuture.ContinueWith(t => LogException(dataForTheFuture)
                , TaskContinuationOptions.OnlyOnFaulted);

            // A call to Wait() or Result will block until the task is completed.
            try
            {
                dataForTheFuture.Wait();
            }
            catch (AggregateException){}    // Ignore, because we do expect them
        }

        private int ComputeSomeResult(int millisecondsWait)
        {
            // Simulate computing some result
            Thread.Sleep(millisecondsWait);
            return 2010;
        }

        private int ComputeEx(int millisecondsWait)
        {
            // Simulate computing some result
            Thread.Sleep(millisecondsWait);
            throw new Exception("ComputeEx");
        }

        private void LogException<T>(Task<T> t)
        {
            Console.WriteLine("Task ex: {0}", t.Exception.Message);
            Console.WriteLine();
        }

    }
}
